/* $Id: eval.y,v 1.1 1994/07/09 19:54:20 ryo Exp $
 *
 *	\[XR[hWFl[^
 *	xt@Cp[T
 *	Copyright (C) 1989,1990 K.Abe
 *	All rights reserved.
 *	Copyright (C) 1997,1998 Tachibana Eriko.
 *
 */

%{
#include <alloca.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "disasm.h"
#include "estruct.h"
#include "etc.h"
#include "fpconv.h"
#include "generate.h"	/* symset */
#include "global.h"
#include "hex.h"	/* strend */
#include "label.h"
#include "labelfile.h"
#include "offset.h"	/* depend_address */
#include "table.h"


#ifndef	uchar
typedef unsigned char	uchar;
#endif

/*
#define	YYDEBUG	1
*/
#ifdef SELF
extern int	yydebug;
#endif

parse_mode	ParseMode;
char		*Lexptr;

int		Eval_Value;
address		Eval_TableTop;
opesize 	Eval_ID;
address		Eval_PC;
int		Eval_Bytes;
char		Eval_ResultString[256];
int		Eval_Break;

int	Eval_Count;

struct expbuf {
    boolean	isvalue;
    boolean	registed;
    address	value;
    char	*bufptr;
};

static int		yylex( void );
static void		yyerror( char* );
static void		tabledesc( int , opesize );
static unsigned long	peek( address , opesize );
static unsigned long	extend( unsigned long , opesize );
static struct expbuf	*storestr( char* );
static struct expbuf	*storeexp( address , int );
static struct expbuf	*store( struct expbuf * );
static void		breakjob( unsigned long );
static void		itoxd_by_size( char* , ULONG , opesize );


#define	EXPSTACKSIZE	32
static struct expbuf	*ExpStack[ EXPSTACKSIZE ];
static int		ExpStackCounter = 0;


%}

/*

	Begin Grammer

*/
%union {
	unsigned long	val;
	char		*str;
	struct expbuf	*expbuf;
}
%token	<str>		_STRING
%token	<val>		SIZEID FSIZEID
%token	<val>		NUMBER
%token	<val>		PEEKB PEEKW PEEKL BREAK EXTW EXTL
%type	<expbuf>	str_or_exp	/* exprstring */
%type	<val>		sizeid fsizeid
%type	<val>		tabledesc
%type	<val>		expr logical_AND_expr equality_expr relational_expr
%type	<val>		additive_expr mul_expr factor element
%left	'<' LE '>'	GE EQUEQU NOTEQU
%left	'(' ')' '{' '}'
%left	','
%left	OROR
%left	ANDAND
%left	'+' '-'
%left	'*' '/'
%left	'!'
%left	UNARYMINUS

%%

tabledesc : BREAK expr				{ breakjob( $2 ); }
	| sizeid				{ tabledesc(  1 , $1 ); }
	| sizeid '[' expr ']'			{ tabledesc( $3 , $1 ); }
	| sizeid exprstring			{ tabledesc(  1 , $1 ); }
	| sizeid '[' expr ']' exprstring	{ tabledesc( $3 , $1 ); }
	| fsizeid				{ tabledesc(  1 , $1 ); }
	| fsizeid '[' expr ']'			{ tabledesc( $3 , $1 ); }
	;

sizeid : SIZEID					{ $$ = $1; }
	| '@' SIZEID				{ $$ = $2; }
	;

fsizeid : FSIZEID				{ $$ = $1; }
	| '@' FSIZEID				{ $$ = $2; }
	;

exprstring : str_or_exp				{ store( $1 ); }
	| exprstring ',' str_or_exp 		{ store( $3 ); }
	;

str_or_exp : _STRING				{ $$ = storestr( $1 ); }
	| expr					{ $$ = storeexp( (address)$1 , 0 ); }
	| '{' expr '}'				{ $$ = storeexp( (address)$2 , 1 ); }
	;

expr : logical_AND_expr				{ $$ = $1; }
	| expr OROR logical_AND_expr		{ $$ = $1 || $3; }
	;

logical_AND_expr : equality_expr		{ $$ = $1; }
	| logical_AND_expr ANDAND equality_expr	{ $$ = $1 && $3; }
	;

equality_expr : relational_expr			{ $$ = $1; }
	| equality_expr EQUEQU relational_expr	{ $$ = ( $1 == $3 ); }
	| equality_expr NOTEQU relational_expr	{ $$ = ( $1 != $3 ); }
	;

relational_expr : additive_expr			{ $$ = $1; }
	| relational_expr '<' additive_expr	{ $$ = ( $1 <  $3 ); }
	| relational_expr LE additive_expr	{ $$ = ( $1 <= $3 ); }
	| relational_expr '>' additive_expr	{ $$ = ( $1 >  $3 ); }
	| relational_expr GE additive_expr	{ $$ = ( $1 >= $3 ); }
	;

additive_expr : mul_expr			{ $$ = $1; }
	| additive_expr '+' mul_expr		{ $$ = $1 + $3; }
	| additive_expr '-' mul_expr		{ $$ = $1 - $3; }
	;

mul_expr : factor				{ $$ = $1; }
	| mul_expr '*' factor			{ $$ = $1 * $3; }
	| mul_expr '/' factor			{ $$ = $1 / $3; }
	| mul_expr '%' factor			{ $$ = $1 % $3; }
	;

factor : '-' factor %prec UNARYMINUS	 	{ $$ = - $2; }
	| '!' factor				{ $$ = ! $2; }
	| element				{ $$ = $1; }
	;

element : '(' expr ')'				{ $$ = $2; }
	| NUMBER				{ $$ = $1; }
	| PEEKB '(' expr ')'	{ $$ = peek( (address)$3 , BYTESIZE ); }
	| PEEKW '(' expr ')'	{ $$ = peek( (address)$3 , WORDSIZE ); }
	| PEEKL '(' expr ')'	{ $$ = peek( (address)$3 , LONGSIZE ); }
	| EXTW '(' expr ')'	{ $$ = extend( (unsigned)$3 , WORDSIZE ); }
	| EXTL '(' expr ')'	{ $$ = extend( (unsigned)$3 , LONGSIZE ); }
	;
%%

/*

	End of Grammer

*/
static int	yylex()
{
    uchar	c;

    while ((c = *Lexptr++) == ' ' || c == (uchar)'\t' || c == (uchar)'\n')
	;
    if( c == (uchar)'\0' )
	return 0;

    if( isdigit(c) ) {
	yylval.val = c - '0';
	while( isdigit( c = *Lexptr++ ) )
	    yylval.val = yylval.val * 10 + c - '0';
	Lexptr--;
	return NUMBER;
    }
    if( c == (uchar)'$' ) {
	yylval.val = 0;
	if( !isxdigit( *Lexptr ) )
	    return 1;
	while( isxdigit( c = *Lexptr++ ) )
	    yylval.val = yylval.val * 16 + ( toupper(c) >= 'A' ? toupper(c)-'A'+10 : c-'0' );
	Lexptr--;
	return NUMBER;
    }
    if( c == (uchar)'"' ) {
	char	*term;

	if( ( term = strchr( Lexptr , '"' ) ) == NULL )
	    return 1;
	yylval.str = Malloc( term - Lexptr + 1 );
	strncpy( yylval.str , Lexptr , term - Lexptr );
	yylval.str[ term - Lexptr ] = '\0';
	Lexptr = term + 1;
	return _STRING;
    }
    if( c == (uchar)'<' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return LE;
	}
	return c;
    }
    if( c == (uchar)'>' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return GE;
	}
	return c;
    }
    if( c == (uchar)'|' ) {
	if( *Lexptr == '|' ) {
	    Lexptr++;
	    return OROR;
	}
	return c;
    }
    if( c == (uchar)'&' ) {
	if( *Lexptr == '&' ) {
	    Lexptr++;
	    return ANDAND;
	}
	return c;
    }
    if( c == (uchar)'=' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return EQUEQU;
	}
	return c;
    }
    if( c == (uchar)'!' ) {
	if( *Lexptr == '=' ) {
	    Lexptr++;
	    return NOTEQU;
	}
	return c;
    }
    if (iscsymf (c) || (c == (uchar)'.')) {
	int	token_len;
	char*	token = alloca (strlen (--Lexptr) + 1);
	char*	p = token;

	/* g[N؂o */
	do
	    *p++ = *Lexptr++;
	while (iscsym (*Lexptr) || *Lexptr == (char)'.');
	*p = '\0';
	token_len = p - token;

	/* SIZEID, FSIZEID */
	{
	    struct idpair {
		char	idstr[8];
		int	idnum;
	    } idbuf[] = {
		{ "dc.b"  , BYTESIZE },
		{ "dc.w"  , WORDSIZE },
		{ "dc.l"  , LONGSIZE },
		{ "dc"	  , WORDSIZE },
		{ "byte"  , BYTEID },
		{ "even"  , EVENID },
		{ "cr"	  , CRID },
		{ "asciz" , ASCIIZID },
		{ "asciiz", ASCIIZID },
		{ "ascii" , ASCIIID },
		{ "lascii", LASCIIID },
#if 0
		{ "long"  , LONGID },
		{ "word"  , WORDID },
#endif
		{ "dc.s"  , SINGLESIZE },
		{ "dc.d"  , DOUBLESIZE },
		{ "dc.x"  , EXTENDSIZE },
		{ "dc.p"  , PACKEDSIZE },
		{ ""	  , 0 }		/* end of list */
	    };
	    struct idpair* idptr = idbuf;

	    if (c == (uchar)'.') token++;
	    for (idptr = idbuf; idptr->idstr[0]; idptr++) {
		if (strcmpi (token, idptr->idstr) == 0) {
		    yylval.val = (unsigned long)idptr->idnum;
		    if (SINGLESIZE <= idptr->idnum)
			return FSIZEID;
		    else
			return SIZEID;
		}
	    }
	    if (c == (uchar)'.') token--;
	}

	if (strcmpi (token, "tabletop") == 0) {
	    yylval.val = (unsigned long)Eval_TableTop;
	    return NUMBER;
	}
	if (strcmpi (token, "pc" ) == 0) {
	    yylval.val = (unsigned long)Eval_PC;
	    return NUMBER;
	}
	if (strcmpi (token, "break") == 0)	return BREAK;
	if (strcmpi (token, "peek.b") == 0)	return PEEKB;
	if (strcmpi (token, "peek.w") == 0)	return PEEKW;
	if (strcmpi (token, "peek.l") == 0)	return PEEKL;
	if (strcmpi (token, "ext.w") == 0)	return EXTW;
	if (strcmpi (token, "ext.l") == 0)	return EXTL;

	/* g[Nł͂Ȃ */
	Lexptr -= token_len;
	/* c = */ Lexptr++;
    }
    return c;
}


static void	yyerror( char *s )
{
    eprintf( "%s\n" , s );
}


static void
fpconv_by_size (char* buf, address ptr, opesize size)
{
    unsigned long val[3];

    val[0] = PEEK_LONG (ptr);
    if (size == SINGLESIZE) {
	fpconv_s (buf, (float*) &val);
	return;
    }

    ptr += 4;
    val[1] = PEEK_LONG (ptr);
    if (size == DOUBLESIZE) {
	fpconv_d (buf, (double*) &val);
	return;
    }

    ptr += 4;
    val[2] = PEEK_LONG (ptr);
    if (size == EXTENDSIZE)
	fpconv_x (buf, (long double*) &val);
    else /* if (size == PACKEDSIZE) */
	fpconv_p (buf, (packed_decimal*) &val);
}


static void	tabledesc( int num , opesize id )
{
    int i;

    Eval_ID = id;
    if( Eval_Count == 0 ) {		/* easy trick ? */
	if (id == ASCIIID || id == BYTEID
#if 0
	 || id == WORDID || id == LONGID
#endif
	) {
	    Eval_Count = 1;
	    Eval_Bytes = num;
	} else
	    Eval_Count = num;
    }


    if( ExpStackCounter == 0 && ParseMode == PARSE_GENERATING ) {
	char* bufptr = strend (Eval_ResultString);

	switch (id) {
	case LONGSIZE:
	    if (depend_address (Eval_PC))
		make_proper_symbol (bufptr, (address)PEEK_LONG (Eval_PC));
	    else
		itoxd_by_size (bufptr, PEEK_LONG (Eval_PC), id);
	    break;
	case WORDSIZE:
	    itoxd_by_size (bufptr, PEEK_WORD (Eval_PC), id);
	    break;
	case BYTESIZE:
	    itoxd_by_size (bufptr, PEEK_BYTE (Eval_PC), id);
	    break;

	case SINGLESIZE:
	case DOUBLESIZE:
	case EXTENDSIZE:
	case PACKEDSIZE:
	    fpconv_by_size (bufptr, Eval_PC, id);
	    break;

	default:	/* reduce warning message */
	    break;
	}
    }
    for( i = 0 ; i < ExpStackCounter ; i++ ) {
	switch( ParseMode ) {
	case PARSE_ANALYZING :
	    if( !( ExpStack[ i ]->isvalue ) )
		free( ExpStack[ i ]->bufptr );
	    free( ExpStack[ i ] );
	    break;
	case PARSE_GENERATING :
	    if( ExpStack[ i ]->isvalue ) {
		char* bufptr = strend (Eval_ResultString);
		address val = ExpStack[ i ]->value;

		if (ExpStack[i]->registed)
		    make_proper_symbol (bufptr, val);
		else
		    itoxd_by_size (bufptr, (ULONG)val, Eval_ID);
	    } else {
		strcat( Eval_ResultString , ExpStack[ i ]->bufptr );
		free( ExpStack[ i ]->bufptr );
	    }
	    free( ExpStack[ i ] );
	    break;
	}
    }
    ExpStackCounter = 0;
}


static struct expbuf	*storestr( char *s )
{
    struct expbuf	*expbufptr;

    expbufptr = Malloc( sizeof( struct expbuf ) );
    expbufptr->bufptr = s;
    expbufptr->isvalue = FALSE;
    return expbufptr;
}


static struct expbuf	*storeexp( address v , int mode )
{
    struct expbuf	*expbufptr;

    expbufptr = Malloc( sizeof( struct expbuf ) );
    expbufptr->value = v;
    expbufptr->isvalue = TRUE;

    if( mode == 1 ) {
	expbufptr->registed = TRUE;
	if( ParseMode == PARSE_ANALYZING && !regist_label( v , DATLABEL | UNKNOWN ) )
	    printf( "??? address %x\n" , (unsigned int)v );
    } else
	expbufptr->registed = FALSE;
    return expbufptr;
}


static struct expbuf	*store( struct expbuf *s )
{
    ExpStack[ ExpStackCounter++ ] = s;
    if( ExpStackCounter == EXPSTACKSIZE )
	err ("]X^bNӂ܂\n");

#if 0	/* debug */
    switch (s->isvalue) {
    case TRUE :
	printf ("l = %d\n", s->value);
	break;
    case FALSE :
	printf (" = %s\n", s->bufptr);
    }
#endif
    return s;
}


static void	itoxd_by_size( char *buf , ULONG num , opesize size )
{
    itoxd (buf, num, (size == LONGSIZE) ? 8 :
		     (size == WORDSIZE) ? 4 : 2);
}


static void
peek_warning (address adrs, int size)
{
    if ((int)adrs & 1)
	eprintf ("Warning: peek.%c ̈AhX(%x)ł\n", size, adrs);
}

static unsigned long
peek (address adrs, opesize size)
{
    unsigned long rc;

    switch (size) {
    case BYTESIZE:
	rc = PEEK_BYTE (adrs);
	break;
    case WORDSIZE:
	peek_warning (adrs, 'w');
	rc = PEEK_WORD (adrs);
	break;
    case LONGSIZE:
	peek_warning (adrs, 'l');
	rc = PEEK_LONG (adrs);
	break;
    default:
	rc = 0;
	break;
    }
    /*
	printf ("peek.%d(%x) = %x\n", size, (unsigned int)adrs, rc);
    */
    return rc;
}



static unsigned long
extend (unsigned long arg, opesize size)
{
    if (size == WORDSIZE)
	return (unsigned short)(signed char)arg;
    else /* if (size == LONGSIZE) */
	return (unsigned long)(signed short)arg;
}


static void	breakjob( unsigned long value )
{
    if( Eval_Count == 0 )			/* easy trick ? */
	Eval_Count = 1;
    Eval_ID = BREAKID;
    Eval_Break = value;
}


#ifdef SELF
main()
{
    char	Buffer[256];

    yydebug = 0;
    while( fgets( Buffer , sizeof Buffer , stdin ) ) {
	Buffer[ strlen( Buffer ) - 1 ] = '\0';
	Lexptr = Buffer;
	yyparse();
    }
}
#endif /* SELF */


/* EOF */
